@@ -2,32 +2,30 @@
 # this repository contains the full copyright notices and license terms.
 import datetime
 import functools
-from decimal import Decimal
 from collections import defaultdict
 from copy import deepcopy
+from decimal import Decimal
 
-from simpleeval import simple_eval
-
+from simpleeval import InvalidExpression, simple_eval
 from sql import Literal, Null, Select, Window, With
 from sql.aggregate import Max, Sum
-from sql.conditionals import Coalesce, Case
+from sql.conditionals import Case, Coalesce
 from sql.functions import CurrentTimestamp
 from sql.operators import Concat
 
 from trytond.i18n import gettext
-from trytond.model import ModelSQL, ModelView, fields
+from trytond.model import Index, ModelSQL, ModelView, fields
 from trytond.model.exceptions import AccessError
-from trytond.wizard import (
-    Wizard, StateTransition, StateAction, StateView, Button)
-from trytond.pyson import Eval, If, Bool, PYSONEncoder
-from trytond.tools import decistmt
-from trytond.transaction import Transaction
+from trytond.modules.product import price_digits, round_price
 from trytond.pool import Pool, PoolMeta
-from trytond.tools import grouped_slice
+from trytond.pyson import Bool, Eval, If, PYSONEncoder
+from trytond.tools import decistmt, grouped_slice
+from trytond.tools import timezone as tz
+from trytond.transaction import Transaction, without_check_access
+from trytond.wizard import (
+    Button, StateAction, StateTransition, StateView, Wizard)
 
-from trytond.modules.product import round_price, price_digits
-
-from .exceptions import ProductCostPriceError
+from .exceptions import ProductCostPriceError, ProductStockWarning
 from .move import StockMixin
 from .shipment import ShipmentAssignMixin
 
@@ -55,8 +53,7 @@ def check_no_move(func):
         pool = Pool()
         Template = pool.get('product.template')
         transaction = Transaction()
-        if (transaction.user != 0
-                and transaction.context.get('_check_access')):
+        if transaction.user and transaction.check_access:
             actions = iter(args)
             for records, values in zip(actions, actions):
                 for field, msg in Template._modify_no_move:
@@ -83,16 +80,87 @@ def check_no_move(func):
     return decorator
 
 
+def check_no_stock_if_inactive(func):
+
+    def get_product_locations(company, location_ids, sub_products):
+        pool = Pool()
+        Product = pool.get('product.product')
+
+        product2locations = defaultdict(list)
+        product_ids = list(map(int, sub_products))
+        with Transaction().set_context(company=company.id):
+            quantities = Product.products_by_location(
+                location_ids, with_childs=True,
+                grouping=('product',), grouping_filter=(product_ids,))
+        for key, quantity in quantities.items():
+            location_id, product_id, = key
+            if quantity:
+                product2locations[product_id].append(location_id)
+        return product2locations
+
+    def raise_warning(company, product2locations):
+        pool = Pool()
+        Location = pool.get('stock.location')
+        Warning = pool.get('res.user.warning')
+        Product = pool.get('product.product')
+
+        for product_id, location_ids in product2locations.items():
+            product = Product(product_id)
+            locations = ','.join(
+                l.rec_name for l in Location.browse(location_ids[:5]))
+            if len(location_ids) > 5:
+                locations += '...'
+            warning_name = Warning.format(
+                'deactivate_product_with_stock', [product])
+            if Warning.check(warning_name):
+                raise ProductStockWarning(warning_name,
+                    gettext(
+                        'stock.msg_product_location_quantity',
+                        product=product.rec_name,
+                        company=company.rec_name,
+                        locations=locations),
+                    gettext('stock.msg_product_location_quantity_description'))
+
+    @functools.wraps(func)
+    def decorator(cls, *args):
+        pool = Pool()
+        Company = pool.get('company.company')
+        Location = pool.get('stock.location')
+
+        to_check = []
+        actions = iter(args)
+        for products, values in zip(actions, actions):
+            if not values.get('active', True):
+                to_check.extend(products)
+        if to_check:
+            with without_check_access():
+                locations = Location.search([('type', '=', 'storage')])
+                location_ids = list(map(int, locations))
+                for company in Company.search([]):
+                    for sub_products in grouped_slice(to_check):
+                        product2locations = get_product_locations(
+                            company, location_ids, sub_products)
+                        raise_warning(company, product2locations)
+        return func(cls, *args)
+    return decorator
+
+
 class Template(metaclass=PoolMeta):
     __name__ = "product.template"
-    quantity = fields.Function(fields.Float('Quantity',
-        help="The amount of stock in the location."),
+    quantity = fields.Function(
+        fields.Float(
+            "Quantity", digits='default_uom',
+            help="The amount of stock in the location."),
         'sum_product')
-    forecast_quantity = fields.Function(fields.Float('Forecast Quantity',
-        help="The amount of stock expected to be in the location."),
+    forecast_quantity = fields.Function(
+        fields.Float(
+            "Forecast Quantity", digits='default_uom',
+            help="The amount of stock expected to be in the location."),
         'sum_product')
-    cost_value = fields.Function(fields.Numeric('Cost Value',
-        help="The value of the stock in the location."),
+    cost_value = fields.Function(
+        fields.Numeric(
+            "Cost Value", digits=price_digits,
+            help="The value of the stock in the location."),
         'sum_product')
 
     def sum_product(self, name):
@@ -129,13 +197,11 @@ class Template(metaclass=PoolMeta):
 class Product(StockMixin, object, metaclass=PoolMeta):
     __name__ = "product.product"
     quantity = fields.Function(fields.Float(
-            "Quantity", digits=(16, Eval('default_uom_digits', 2)),
-            depends=['default_uom_digits'],
+            "Quantity", digits='default_uom',
             help="The amount of stock in the location."),
         'get_quantity', searcher='search_quantity')
     forecast_quantity = fields.Function(fields.Float(
-            "Forecast Quantity", digits=(16, Eval('default_uom_digits', 2)),
-            depends=['default_uom_digits'],
+            "Forecast Quantity", digits='default_uom',
             help="The amount of stock expected to be in the location."),
         'get_quantity', searcher='search_quantity')
     cost_value = fields.Function(fields.Numeric(
@@ -157,6 +223,8 @@ class Product(StockMixin, object, metaclass=PoolMeta):
 
     @classmethod
     def get_cost_value(cls, products, name):
+        pool = Pool()
+        Company = pool.get('company.company')
         cost_values = {p.id: None for p in products}
         context = {}
         trans_context = Transaction().context
@@ -164,6 +232,19 @@ class Product(StockMixin, object, metaclass=PoolMeta):
             # Use the last cost_price of the day
             context['_datetime'] = datetime.datetime.combine(
                 trans_context['stock_date_end'], datetime.time.max)
+            company = trans_context.get('company')
+            if company:
+                company = Company(company)
+                if company.timezone:
+                    timezone = tz.ZoneInfo(company.timezone)
+                    try:
+                        context['_datetime'] = (
+                            context['_datetime']
+                            .replace(tzinfo=timezone)
+                            .astimezone(tz.UTC)
+                            .replace(tzinfo=None))
+                    except OverflowError:
+                        pass
             # The date could be before the product creation
             products = [p for p in products
                 if p.create_date <= context['_datetime']]
@@ -177,6 +258,7 @@ class Product(StockMixin, object, metaclass=PoolMeta):
 
     @classmethod
     @check_no_move
+    @check_no_stock_if_inactive
     def write(cls, *args):
         super(Product, cls).write(*args)
 
@@ -277,15 +359,14 @@ class Product(StockMixin, object, metaclass=PoolMeta):
             cls.update_cost_price(costs)
 
     @classmethod
+    @without_check_access
     def update_cost_price(cls, costs):
         "Update cost price of products from costs re-computation dictionary"
         to_write = []
         for cost, products in costs.items():
             to_write.append(products)
             to_write.append({'cost_price': cost})
-
-        with Transaction().set_context(_check_access=False):
-            cls.write(*to_write)
+        cls.write(*to_write)
 
     def recompute_cost_price_fixed(self, start=None):
         return self.cost_price or Decimal(0)
@@ -385,8 +466,14 @@ class Product(StockMixin, object, metaclass=PoolMeta):
         def out_move(move):
             return not in_move(move)
 
+        def production_move(move):
+            return (
+                move.from_location.type == 'production'
+                or move.to_location.type == 'production')
+
         current_moves = []
         current_cost_price = cost_price
+        qty_production = 0
         for move in moves:
             if (current_moves
                     and current_moves[-1].effective_date
@@ -396,24 +483,31 @@ class Product(StockMixin, object, metaclass=PoolMeta):
                         if m.cost_price != current_cost_price],
                     dict(cost_price=current_cost_price))
                 current_moves.clear()
+                qty_production = 0
             current_moves.append(move)
 
             cost_price = Revision.apply_up_to(
                 revisions, cost_price, move.effective_date)
-            qty = Uom.compute_qty(move.uom, move.quantity, self.default_uom)
+            qty = Uom.compute_qty(move.unit, move.quantity, self.default_uom)
             qty = Decimal(str(qty))
             if out_move(move):
                 qty *= -1
             if in_move(move):
+                in_qty = qty
+                if production_move(move) and qty_production < 0:
+                    # Exclude quantity coming back from production
+                    in_qty -= min(abs(qty_production), in_qty)
                 unit_price = move.get_cost_price(product_cost_price=cost_price)
-                if quantity + qty > 0 and quantity >= 0:
+                if quantity + in_qty > 0 and quantity >= 0:
                     cost_price = (
-                        (cost_price * quantity) + (unit_price * qty)
-                        ) / (quantity + qty)
-                elif qty > 0:
+                        (cost_price * quantity) + (unit_price * in_qty)
+                        ) / (quantity + in_qty)
+                elif in_qty > 0:
                     cost_price = unit_price
                 current_cost_price = round_price(cost_price)
             quantity += qty
+            if production_move(move):
+                qty_production += qty
 
         Move.write([
                 m for m in current_moves
@@ -437,6 +531,8 @@ class Product(StockMixin, object, metaclass=PoolMeta):
 class ProductByLocationContext(ModelView):
     'Product by Location'
     __name__ = 'product.by_location.context'
+
+    company = fields.Many2One('company.company', "Company", required=True)
     forecast_date = fields.Date(
         'At Date',
         help="The date for which the stock quantity is calculated.\n"
@@ -445,6 +541,10 @@ class ProductByLocationContext(ModelView):
     stock_date_end = fields.Function(fields.Date('At Date'),
         'on_change_with_stock_date_end')
 
+    @classmethod
+    def default_company(cls):
+        return Transaction().context.get('company')
+
     @staticmethod
     def default_forecast_date():
         Date = Pool().get('ir.date')
@@ -514,7 +614,7 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
                 date = v[name]
                 # SQLite does not convert to date
                 if isinstance(date, str):
-                    date = datetime.date(*map(int, date.split('-', 2)))
+                    date = datetime.date.fromisoformat(date)
                 result[v['id']] = date
             return result
 
@@ -679,7 +779,7 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
 
         def cast_date(date):
             if isinstance(date, str):
-                date = datetime.date(*map(int, date.split('-', 2)))
+                date = datetime.date.fromisoformat(date)
             return date
 
         dates = sorted({cast_date(l.date) for l in lines})
@@ -721,6 +821,8 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
 class ProductQuantitiesByWarehouseContext(ModelView):
     'Product Quantities By Warehouse'
     __name__ = 'stock.product_quantities_warehouse.context'
+
+    company = fields.Many2One('company.company', "Company", required=True)
     warehouse = fields.Many2One('stock.location', 'Warehouse', required=True,
         domain=[
             ('type', '=', 'warehouse'),
@@ -730,6 +832,10 @@ class ProductQuantitiesByWarehouseContext(ModelView):
         "Only storage zone",
         help="Check to use only the quantity of the storage zone.")
 
+    @classmethod
+    def default_company(cls):
+        return Transaction().context.get('company')
+
     @classmethod
     def default_warehouse(cls):
         Location = Pool().get('stock.location')
@@ -917,7 +1023,7 @@ class ProductQuantitiesByWarehouseMove(ModelSQL, ModelView):
 
         def cast_date(date):
             if isinstance(date, str):
-                date = datetime.date(*map(int, date.split('-', 2)))
+                date = datetime.date.fromisoformat(date)
             return date
 
         dates = sorted({cast_date(r.date) for r in records})
@@ -1084,6 +1190,8 @@ class ModifyCostPrice(Wizard):
 
     def get_revision(self, Revision):
         return Revision(
+            template=None,
+            product=None,
             company=Transaction().context.get('company'),
             date=self.start.date,
             cost_price=self.start.cost_price,
@@ -1115,7 +1223,7 @@ class CostPriceRevision(ModelSQL, ModifyCostPriceStart):
     __name__ = 'product.cost_price.revision'
     template = fields.Many2One(
         'product.template', "Product",
-        ondelete='CASCADE', select=True, required=True,
+        ondelete='CASCADE', required=True,
         domain=[
             If(Bool(Eval('product')),
                 ('products', '=', Eval('product')),
@@ -1124,10 +1232,10 @@ class CostPriceRevision(ModelSQL, ModifyCostPriceStart):
         context={
             'company': Eval('company', -1),
             },
-        depends=['product', 'company'])
+        depends={'company'})
     product = fields.Many2One(
         'product.product', "Variant",
-        ondelete='CASCADE', select=True,
+        ondelete='CASCADE',
         domain=[
             If(Bool(Eval('template')),
                 ('template', '=', Eval('template')),
@@ -1136,13 +1244,20 @@ class CostPriceRevision(ModelSQL, ModifyCostPriceStart):
         context={
             'company': Eval('company', -1),
             },
-        depends=['template', 'company'])
+        depends={'company'})
     company = fields.Many2One(
         'company.company', "Company", ondelete='CASCADE', required=True)
 
     @classmethod
     def __setup__(cls):
         super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.product, Index.Equality()),
+                (t.template, Index.Equality()),
+                (t.company, Index.Equality())))
         cls._order.insert(0, ('date', 'DESC'))
 
     @classmethod
@@ -1155,26 +1270,36 @@ class CostPriceRevision(ModelSQL, ModifyCostPriceStart):
             self.template = self.product.template
 
     @classmethod
-    def validate(cls, revisions):
-        super().validate(revisions)
-        for revision in revisions:
-            revision.check_cost_price()
+    def validate_fields(cls, revisions, field_names):
+        super().validate_fields(revisions, field_names)
+        cls.check_cost_price(revisions, field_names)
 
-    def check_cost_price(self):
-        try:
-            if not isinstance(self.get_cost_price(Decimal(0)), Decimal):
-                raise ValueError
-        except Exception as exception:
-            raise ProductCostPriceError(
-                gettext('stock.msg_invalid_cost_price',
-                    cost_price=self.cost_price,
-                    product=(self.product or self.template).rec_name,
-                    exception=exception)) from exception
+    @classmethod
+    def check_cost_price(cls, revisions, field_names):
+        if field_names and 'cost_price' not in field_names:
+            return
+        for revision in revisions:
+            revision.get_cost_price(Decimal(0))
 
     def get_cost_price(self, cost_price, **context):
         context.setdefault('names', {})['cost_price'] = cost_price
         context.setdefault('functions', {})['Decimal'] = Decimal
-        return simple_eval(decistmt(self.cost_price), **context)
+        try:
+            amount = simple_eval(decistmt(self.cost_price), **context)
+        except (InvalidExpression, SyntaxError) as exception:
+            product = self.product or self.template
+            raise ProductCostPriceError(
+                gettext('stock.msg_invalid_cost_price',
+                    cost_price=self.cost_price,
+                    product=product.rec_name if product else '',
+                    exception=exception)) from exception
+        if not isinstance(amount, Decimal):
+            raise ProductCostPriceError(
+                gettext('stock.msg_invalid_cost_price_not_number',
+                    value=amount,
+                    cost_price=self.cost_price,
+                    product=product.rec_name))
+        return amount
 
     @classmethod
     def _get_for_product_domain(cls):
